# -*- coding: utf-8 -*-

#######################################################################
# Copyright © 2007 Yuri Malheiros.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# the Free Software Foundation; version 2 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#######################################################################

"""
This module implements a class for responsible for manage all
opened text files of TextFlow.
"""

import gtk
import gtksourceview2
import gnomevfs
import gobject
import os
import pango
import gettext

from TextFlow.widgets.TFSourceView import TFSourceView
from TextFlow.core.PreferencesManager import PreferencesManager
from TextFlow.core.TriggerManager import TriggerManager
from TextFlow.com.Search import Search
from TextFlow.core import constants

from TextFlow.widgets.FileDialog import SaveFileDialog as SaveFileDialog
from TextFlow.widgets.CloseAlertDialog import CloseAlertDialog
from TextFlow.widgets.ErrorDialog import ErrorDialog
from TextFlow.widgets.ChangeEncodingDialog import ChangeEncodingDialog

_ = gettext.gettext

class DocumentManager(gtk.Notebook):
    """
    This class extend the gtk.Notebook widget with functions
    to manipulate opened files.
    """

    def __init__(self, interface):
        """
        Constructor.
        """
        super(DocumentManager, self).__init__()
        
        self.interface = interface
        self.main_window = interface.main_window
        
        self.preferences_manager = PreferencesManager()
        self.trigger_manager = TriggerManager(self)
        
        self.search_functions = Search(self)
        
        close_button_style = ''' 
            style 'close_button' {
                xthickness = 0
                ythickness = 0
            }
            widget '*.close_button' style 'close_button'
        '''
        gtk.rc_parse_string(close_button_style)
        
        self.set_scrollable(True)
        self.connect("switch-page", self.change_tab)
        
        self.__get_color_styles()
    
    #################### Public Methods ####################
    
    def open_tab(self, text, file_uri="", encode=""):
        """
        Open a new tab in the DocumentManager
        
        @param text: Tab label text.
        @type text: A string.
        
        @param file_uri: A file that will be displayed in TFSourceView.
        @type file_uri: A string.
        """
        tab_box = gtk.HBox(False, 4)
        tab_label = gtk.Label(text)
        tab_button = gtk.Button()
        tab_button.set_name('tab.close_button')
        
        self.__add_icon_to_button(tab_button)
        
        tab_box.pack_start(tab_label, True, True, 0)
        tab_box.pack_start(tab_button, False, False, 0)
        tab_box.show_all()
        
        swindow = gtk.ScrolledWindow()
        swindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        
        try:
            view = self.__create_view(file_uri, encode)
        except UnicodeDecodeError:
            dialog_encoding = ChangeEncodingDialog()
            button_clicked = dialog_encoding.run()
            
            if button_clicked == gtk.RESPONSE_DELETE_EVENT:
                dialog_encoding.destroy()
                return
            elif button_clicked == 1:
                encoding = dialog_encoding.get_encoding()
                self.open_tab(text, file_uri, encoding)
                dialog_encoding.destroy()
                return
            elif button_clicked == 2:
                dialog_encoding.destroy()
                return
        else:
            swindow.add(view)
            swindow.show_all()
            
            child = self.get_nth_page(0)
            
            if child != None:
                v_child = child.get_child()
            
            if self.get_n_pages() == 1:
                if v_child.get_file_uri() == "" and v_child.get_text() == "":
                    if view.get_file_uri() != "":
                        self.__remove_tab(view, 0)
            elif self.get_n_pages() == 0:
                self.interface.statusbar.show_info()
            
            new_page_num = self.page_num(swindow)
            
            if file_uri != "":
                if self.preferences_manager.get_value("open_save/auto_save"):
                    self.add_to_autosave(view)
            
            self.append_page(swindow, tab_box)
            self.set_tab_reorderable(swindow, True)
            self.set_current_page(new_page_num)
            tab_button.connect('clicked', self.close_tab_button, swindow)
            view.buffer.connect('changed', self.buffer_change, view, tab_label)
            view.grab_focus()

    def close_tab(self, swindow):
        """
        Close a tab in the DocumentManager
        
        @param swindow: A ScrolledWindow.
        @type swindow: A ScrolledWindow object.
        """
        view = swindow.get_child()
        
        if not view.is_updated():
            dialog = CloseAlertDialog()
            
            button_clicked = dialog.run()
            
            if button_clicked == gtk.RESPONSE_DELETE_EVENT:
                dialog.destroy()
                return
            elif button_clicked == 1:
                dialog.destroy()
            elif button_clicked == 2:
                dialog.destroy()
                return
            elif button_clicked == 3:
                file_uri = view.get_file_uri()
                if file_uri != "":
                    self.save_file_tab(file_uri)
                else:
                    file_save = SaveFileDialog(constants.MESSAGE_0002, None,
                                               "", "*", False)
                    files = file_save.run()
                    file_save.destroy()
            
                    if len(files):
                        self.save_file_tab(files[0])
                
                dialog.destroy()
            
        if view.file_uri != "" and view.autosave:
            self.remove_from_autosave(view)
        
        #Close tab frees memory correctly?
        page = self.page_num(swindow)
        self.__remove_tab(view, page)
        
        if self.get_n_pages() == 0:
            self.main_window.set_title(constants.MESSAGE_0015)
            #self.interface.statusbar.clear()
            self.interface.statusbar.hide_info()
        
    def open_file_tab(self, new_file, encode=""):
        """
        Open a new tab with a file associated.
        
        @param new_file: A file that will be displayed in the new tab.
        @type new_file: A string.
        """
        self.open_tab(os.path.basename(new_file), new_file, encode)
    
    def save_file_tab(self, save_file, show_error=True):
        """
        Save the text of active tab in a file.
        
        @param save_file: A file that will be save.
        @type save_file: A string.
        """
        view = self.get_active_view()
        self.__save_view(view, save_file, show_error)
    
    def get_active_view(self):
        """
        Get the active TFSourceView
        
        @return: The active TFSourceView.
        @rtype: A TFSourceView object.
        """
        current_page = self.get_current_page()
        if current_page == -1:
            return None
        else:
            return self.get_nth_page(current_page).get_child()
    
    def store_tabs(self):
        """
        Save file uri and the encode of all tabs in a text file.
        """
        store_file_uri = constants.HOME + "/.textflow/stored_tabs"
        store_file = open(store_file_uri, "w")
        
        num_pages = self.get_n_pages()
        
        for i in range(num_pages):
            view = self.get_nth_page(i).get_child()
            if view.file_uri != "":
                store_file.write(view.file_uri + "\n" + view.encode + "\n")
        
        store_file.close()
        
    def open_stored_tabs(self):
        """
        Open tabs according stored_tabs file data.
        """
        
        store_file_uri = constants.HOME + "/.textflow/stored_tabs"
        try:
            store_file = open(store_file_uri, "r")
        except IOError:
            return
        else:
            lines = store_file.readlines()
            
            cont = 0
            while cont < len(lines):
                if os.path.exists(lines[cont][:-1]):
                    self.open_file_tab(lines[cont][:-1], lines[cont+1][:-1])
                
                cont += 2
            
    def change_encode(self, encode):
        page_num = self.get_current_page()
        
        if page_num != -1:
            swindow = self.get_nth_page(page_num)
            view = swindow.get_child()

            if view.encode != encode:
                file_uri = view.file_uri
                
                if file_uri != "":
                    if not view.is_updated():
                        dialog = CloseAlertDialog()
                        dialog.set_markup(constants.MESSAGE_0014)

                        button_clicked = dialog.run()

                        if button_clicked == gtk.RESPONSE_DELETE_EVENT:
                            dialog.destroy()
                            return
                        elif button_clicked == 1:
                            dialog.destroy()
                        elif button_clicked == 2:
                            dialog.destroy()
                            return
                        elif button_clicked == 3:
                            self.save_file_tab(file_uri)
                           
                            dialog.destroy()
                                                       
                        page = self.page_num(swindow)
                        self.__remove_tab(view, page)
                        self.open_file_tab(file_uri, encode)
                    else:
                        page = self.page_num(swindow)
                        self.__remove_tab(view, page)
                        self.open_file_tab(file_uri, encode)
                        
                    new_page_num = self.get_current_page()
                    swindow = self.get_nth_page(new_page_num)
                    self.reorder_child(swindow, page_num)
    
    def add_to_autosave(self, view):
        minutes = self.preferences_manager.get_value("open_save/auto_save_minutes")
        view.autosave = gobject.timeout_add(minutes*60000, self.__save_view, view,
                                            view.file_uri, priority=gobject.PRIORITY_LOW)

    def remove_from_autosave(self, view):
        gobject.source_remove(view.autosave)
        view.autosave = False

    def set_style(self, view, value):
        """
        Set the color style used by a TFSourceView.
        
        @param view: The TFSourceView.
        @type view: A TFSourceView object.
        
        @param value: The font name.
        @type value: A string.
        """
        
        scheme = self.style_manager.get_scheme(value)
        buffer = view.buffer
        
        if scheme != None:
            buffer.set_style_scheme(scheme)
            
            style_scheme = buffer.get_style_scheme()
            style = style_scheme.get_style("search-match")
            style_sticky = style_scheme.get_style("def:note")
            
            # Set search match color
            if style != None:
                foreground = style.get_property("foreground")
                background = style.get_property("background")
            else:
                foreground = "black"
                background = "lightblue"

            if style_sticky != None:
                foreground_sticky = style_sticky.get_property("foreground")
                background_sticky = style_sticky.get_property("background")
            else:
                foreground_sticky = "black"
                background_sticky = "#ff6100"
                
            tag_table = buffer.get_tag_table()
            tag = tag_table.lookup("searchmatch")
            tag_sticky = tag_table.lookup("sticky")
            
            if foreground != None:
                tag.set_property("foreground", foreground)
                tag_sticky.set_property("foreground", foreground_sticky)
            else:
                tag.set_property("foreground", "black")
                tag_sticky.set_property("foreground", "black")
            
            if background != None:
                tag.set_property("background", background)
                tag_sticky.set_property("background", background_sticky)
            
    def set_text_wrapping(self, view, value):
        """
        Enable or disable text wrapping in a TFSourceView.
        
        @param view: The TFSourceView.
        @type view: A TFSourceView object.
        
        @param value: True to enable text wrapping and False to disable.
        @type value: A boolean.
        """
        if value:
            view.set_wrap_mode(gtk.WRAP_WORD)
        else:
            view.set_wrap_mode(gtk.WRAP_NONE)
        
    #################### Private Methods ####################
    
    def __get_color_styles(self):
        """
        This method gets all available color styles.
        """
        styles_dir = os.getenv("HOME") + "/.textflow/colors"
        self.style_manager = gtksourceview2.StyleSchemeManager()
        self.style_manager.prepend_search_path(styles_dir)
    
    def __add_icon_to_button(self, button):
        """
        Add the close image to a button.
        
        @param button: Reference to a button.
        @type file_uri: A Button object.
        """
        icon_box = gtk.HBox(False, 0)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_CLOSE,gtk.ICON_SIZE_MENU)
        gtk.Button.set_relief(button, gtk.RELIEF_NONE)
        settings = gtk.Widget.get_settings(button)
        (w, h) = gtk.icon_size_lookup_for_settings(settings, gtk.ICON_SIZE_MENU)
        gtk.Widget.set_size_request(button, w + 0, h + 1)
        image.show()
        icon_box.pack_start(image, True, False, 0)
        button.add(icon_box)

        icon_box.show()
    
    def __create_view(self, file_uri, encode):
        """
        Create a TFSourceView and set some properties.
        
        @param file_uri: A file that will be displayed in TFSourceView.
        @type file_uri: A string
        
        @return: The TFSourceView created.
        @rtype: A TFSourceView object.
        """
        buffer = gtksourceview2.Buffer()
        pm = self.preferences_manager
        
        #Color stuff
        buffer.create_tag("searchmatch", background="lightblue")
        buffer.create_tag("sticky", background="#ff6100")
        buffer.set_highlight_syntax(True)
        
        view = TFSourceView(buffer, file_uri, encode)
        
        # Define view encode to default encoding 
        if view.encode == "":
            view.encode = pm.get_value("open_save/encoding")
        
        #applying properties
        #self.set_font(view, pm.get_value("font"))
        view.modify_font(pango.FontDescription(pm.get_value("font")))
        #self.set_tab_width(view, pm.get_value("indentation/tab_width"))
        view.set_tab_width(pm.get_value("indentation/tab_width"))
        #self.set_use_spaces(view, pm.get_value("indentation/use_spaces"))
        view.set_insert_spaces_instead_of_tabs(pm.get_value("indentation/use_spaces"))
        #self.set_auto_indent(view, pm.get_value("indentation/automatic"))
        view.set_auto_indent(pm.get_value("indentation/automatic"))
        #self.set_show_right_margin(view, pm.get_value("right_margin/display"))
        view.set_show_right_margin(pm.get_value("right_margin/display"))
        #self.set_right_margin_position(view, pm.get_value("right_margin/position"))
        view.set_right_margin_position(pm.get_value("right_margin/position"))
        #self.set_show_line_numbers(view, pm.get_value("line_numbers"))
        view.set_show_line_numbers(pm.get_value("line_numbers"))
        #self.set_highlight_current_line(view, pm.get_value("highlight_current_line"))
        view.set_highlight_current_line(pm.get_value("highlight_current_line"))
        self.set_text_wrapping(view, pm.get_value("text_wrapping"))
        #self.set_brackets_matching(view, pm.get_value("brackets_matching"))
        view.buffer.set_highlight_matching_brackets(pm.get_value("brackets_matching"))
        view.set_smart_home_end(False)
        self.set_style(view, pm.get_value("color_style"))
        
        view.connect('key-press-event', self.pre_key_press)
        view.connect('key-release-event', self.key_release)
        view.connect("drag-motion", self.drag_motion)
        view.connect("drag-drop", self.drag_drop)
        view.connect("drag-data-received", self.drag_data_received)
        view.drag_dest_set(gtk.DEST_DEFAULT_ALL, [("text/uri-list", 0, 1)],
                           gtk.gdk.ACTION_COPY)
        view.buffer.connect("notify::cursor-position", self.cursor_moved, view,
                            self.interface.statusbar)
                            
        view.show()
        return view
    
    def __remove_tab(self, view, page):
        view.word_complete.stop_indexer()
        self.remove_page(page)
    
    def __save_view(self, view, save_file, show_error=True):
        
        # Sometimes is impossible to encode the text with current encode
        # TODO: try - catch here! (UnicodeEncodeError)
        try:
            text = view.get_text().encode(view.encode)
        except UnicodeEncodeError:
            error_dialog = ErrorDialog(constants.MESSAGE_0018,
                                       constants.MESSAGE_0025)
            button_clicked = error_dialog.run()
            
            if button_clicked == gtk.RESPONSE_DELETE_EVENT:
                error_dialog.destroy()
                return
            elif button_clicked == 1:
                error_dialog.destroy()
            return
        
        try:
            arquivo = open(save_file, "w")
        except IOError:
            if show_error:
                view.permission = "Read Only"
                self.interface.statusbar.permission_label.set_text(view.permission)
                error_dialog = ErrorDialog(constants.MESSAGE_0018,
                                           constants.MESSAGE_0019)
                button_clicked = error_dialog.run()
                
                if button_clicked == gtk.RESPONSE_DELETE_EVENT:
                    error_dialog.destroy()
                    return
                elif button_clicked == 1:
                    error_dialog.destroy()
            else:
                return
        else:
            
            
            line_ending = self.preferences_manager.get_value("open_save/line_ending")
            
            if line_ending == "CR LF":
                text = text.replace("\n","\r\n")

            if  self.preferences_manager.get_value("open_save/save_copy") == True:
                arquivo_copy = open(save_file + '~', "w")
                arquivo_copy.write(text)                
                arquivo_copy.close()
            
            arquivo.write(text)
            arquivo.close()
            
            if view.file_uri == "":
                if self.preferences_manager.get_value("open_save/auto_save"):
                    self.add_to_autosave(view)
            
            if view.get_file_uri() != save_file:
                view.set_file_uri(save_file)
            
            current_page = self.get_current_page()
            current_child = self.get_nth_page(current_page)
            tab_label = self.get_tab_label(current_child)
            children = tab_label.get_children()
            
            f_list = save_file.split("/")
            file_name = f_list.pop()
            direc = " (" + "/".join(f_list) + ")"
            
            children[0].set_label(file_name)
            
            self.main_window.set_title(str(file_name) + direc)
    #        begin = view.buffer.get_start_iter()
    #        end = view.buffer.get_end_iter()
    #         self.search_functions.update_matches(begin, end)
            
            view.modified = False
            view.permission = "Writable"
            self.interface.statusbar.permission_label.set_text(view.permission)

    #################### Signals ####################

    def close_tab_button(self, widget, swindow):
        """
        Close a tab in the DocumentManager
        
        @param widget: Reference to the widget.
        @type widget: A widget object.
        
        @param swindow: A ScrolledWindow.
        @type swindow: A ScrolledWindow object.
        """
        self.close_tab(swindow)
    
    def change_tab(self, widget, page, page_num):
        """
        This method does some operations when tabs are changed.
        """
        previous_view = widget.get_active_view()
        if previous_view != None:
            previous_view.word_complete.stop_indexer()
            
        view = self.get_nth_page(page_num).get_child()
        
        view.word_complete.start_indexer()
        
        self.search_functions.view = view
        self.search_functions.buffer = view.buffer
        
        self.search_functions.cursor_mark = \
        self.search_functions.buffer.get_insert()
         
        title = view.get_file_uri()
        if title == "":
            file_name = constants.MESSAGE_0001
            direc = ""
        else:
            f_list = title.split("/")
            file_name = f_list.pop()
            direc = " (" + "/".join(f_list) + ")"
        
        self.main_window.set_title(str(file_name) + direc)
        begin = view.buffer.get_start_iter()
        end = view.buffer.get_end_iter()
        self.search_functions.update_matches(begin, end)
        
        pos = view.get_cursor_position()
        self.interface.statusbar.set_line_column(pos[0], pos[1])
        self.interface.statusbar.combobutton_encoding.set_label(view.encode)
        self.interface.statusbar.permission_label.set_text(view.permission)
        #self.interface.statusbar.encoding_label.set_text(view.encode)
    
    def drag_motion(self, textview, context, x, y, time):
        """
        Handles callback when a drag operation begins.
        """
        if "text/uri-list" in context.targets:
            return True
        else:
            return False

    def drag_drop(self, textview, context, x, y, time):
        """
        Handles callback when a drop operation begins.
        """
        if "text/uri-list" in context.targets:
            return True
        else:
            return False
        
    def drag_data_received(self, textview, context, xcord, ycord,
                              selection_data, info, timestamp):
        """
        Handles callback when a drop operation finishes.
        """
        #from operator import contains, not_, ne
        #if not_(contains(context.targets, "text/uri-list")): return False
        if not ("text/uri-list" in context.targets):
            return False
        else:
            if info != 1:
                return False
            
            # Load file
            uri_list = list(selection_data.get_uris())
            
            for i in uri_list:
                self.open_file_tab(gnomevfs.get_local_path_from_uri(i))
            
            context.finish(True, False, timestamp)
            return True                         
    
    def cursor_moved(self, buffer, position, view, statusbar):
        pos = view.get_cursor_position()
        statusbar.set_line_column(pos[0], pos[1])
    
    def key_release(self, widget, event):
        """
        This method does some operations when a key is realeased.
        """
        last = self.trigger_manager.last_shortcut
#         print "--------------------------"
#         print event.state
#         print event.keyval
#         print "--------------------------"
        if last in self.trigger_manager.sticky_keys:
            if event.state & gtk.gdk.CONTROL_MASK and event.state & gtk.gdk.SHIFT_MASK:
                pass
#                 print "ctrl+shift+" + unichr(event.keyval).lower()
            elif event.state & gtk.gdk.CONTROL_MASK:
                if "ctrl" in last and (event.keyval == 65507 or not (unichr(event.keyval) in self.trigger_manager.sticky_keys[last])):
                    try:
                        self.trigger_manager.sticky_shortcuts[last]()
                    #except KeyError:
                    #    return
                    finally:
                        self.trigger_manager.last_shortcut = None
    #         elif event.state & gtk.gdk.SHIFT_MASK:
    #             shortcut = "shift+" + unichr(event.keyval).lower()
    #             shortcut2 = unichr(event.keyval).lower()
    #             if shortcut in self.trigger_manager.sticky_shortcuts:
    #                 self.trigger_manager.sticky_shortcuts[shortcut]()
    #             elif shortcut2 in self.trigger_manager.sticky_shortcuts:
    #                 self.trigger_manager.sticky_shortcuts[shortcut2]()
            else:
                shortcut = unichr(event.keyval).lower()
                last_key = last.split("+")[-1]
                if shortcut == last_key:
                    try:
                        self.trigger_manager.sticky_shortcuts[last]()
                    #except KeyError:
                    #    return
                    finally:
                        self.trigger_manager.last_shortcut = None
        
    def pre_key_press(self, widget, event):
#        print event.keyval
#        print event.state

        if event.state & gtk.gdk.CONTROL_MASK \
        and event.state & gtk.gdk.SHIFT_MASK:
            shortcut = "ctrl+shift+" + unichr(event.keyval).lower()
            if shortcut in self.trigger_manager.shortcuts:
                r = self.trigger_manager.shortcuts[shortcut]()
                self.trigger_manager.last_shortcut = shortcut
                return r
            else:
                return False
        elif event.state & gtk.gdk.CONTROL_MASK:
            shortcut = "ctrl+" + unichr(event.keyval).lower()
            if shortcut in self.trigger_manager.shortcuts:
                r = self.trigger_manager.shortcuts[shortcut]()
                self.trigger_manager.last_shortcut = shortcut
                return r
            else:
                return False
        elif event.state & gtk.gdk.SHIFT_MASK:
            shortcut = "shift+" + unichr(event.keyval).lower()
            shortcut2 = unichr(event.keyval).lower()
            if shortcut in self.trigger_manager.shortcuts:
                r = self.trigger_manager.shortcuts[shortcut]()
                self.trigger_manager.last_shortcut = shortcut
                return r
            elif shortcut2 in self.trigger_manager.shortcuts:
                r = self.trigger_manager.shortcuts[shortcut2]()
                self.trigger_manager.last_shortcut = shortcut2
                return r
            else:
                return False
        else:
            shortcut = unichr(event.keyval).lower()
            if shortcut in self.trigger_manager.shortcuts:
                r = self.trigger_manager.shortcuts[shortcut]()
                self.trigger_manager.last_shortcut = shortcut
                return r
            else:
                return False
        
    def buffer_change(self, widget, view, label):
        if not view.modified:
            text = label.get_text()
            label.set_text("*" + text)
            view.modified = True
        
        self.search_functions.all_search = False
        start, end = view.get_line_iters()
        self.search_functions.update_matches(start, end)
